# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2026 Valory AG
#   Copyright 2018-2021 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This test module contains the tests for CLI fingerprint command."""

from pathlib import Path
from unittest import TestCase, mock
from unittest.mock import MagicMock

import pytest
from click import ClickException

from aea.cli import cli
from aea.cli.fingerprint import fingerprint_item
from aea.cli.utils.context import Context
from aea.cli.utils.decorators import _check_aea_project
from aea.configurations.constants import (
    DEFAULT_CONNECTION_CONFIG_FILE,
    DEFAULT_SKILL_CONFIG_FILE,
)
from aea.test_tools.test_cases import AEATestCaseEmpty

from tests.conftest import CLI_LOG_OPTION, CliRunner
from tests.test_cli.tools_for_testing import ConfigLoaderMock, ContextMock, PublicIdMock


@mock.patch("aea.cli.fingerprint.fingerprint_package")
class FingerprintCommandTestCase(TestCase):
    """Test case for CLI fingerprint command."""

    def setUp(self):
        """Set it up."""
        self.runner = CliRunner()

    def test_fingerprint_positive(self, *mocks):
        """Test for CLI fingerprint positive result."""
        public_id = "author/name:0.1.0"
        result = self.runner.invoke(
            cli,
            [*CLI_LOG_OPTION, "fingerprint", "connection", public_id],
            standalone_mode=False,
        )
        self.assertEqual(result.exit_code, 0)

        result = self.runner.invoke(
            cli,
            [*CLI_LOG_OPTION, "fingerprint", "contract", public_id],
            standalone_mode=False,
        )
        self.assertEqual(result.exit_code, 0)

        result = self.runner.invoke(
            cli,
            [*CLI_LOG_OPTION, "fingerprint", "protocol", public_id],
            standalone_mode=False,
        )
        self.assertEqual(result.exit_code, 0)

        result = self.runner.invoke(
            cli,
            [*CLI_LOG_OPTION, "fingerprint", "skill", public_id],
            standalone_mode=False,
        )
        self.assertEqual(result.exit_code, 0)

    def _run_fingerprint_by_path(self):
        """Call fingerprint by-path cli command."""
        result = self.runner.invoke(
            cli,
            [*CLI_LOG_OPTION, "fingerprint", "by-path", "some_dir"],
            standalone_mode=False,
            catch_exceptions=False,
        )
        self.assertEqual(result.exit_code, 0, result.exception)

    def test_by_path_ok(self, fingerprint_mock):
        """Test fingerprint by_path works ok."""
        with mock.patch("os.listdir", return_value=[DEFAULT_CONNECTION_CONFIG_FILE]):
            self._run_fingerprint_by_path()
            fingerprint_mock.assert_called()

    def test_by_path_exceptions(self, *mocks):
        """Test fingerprint by_path works raises exceptions."""
        with pytest.raises(
            ClickException,
            match="No package config file found in `.*`. Incorrect directory?",
        ):
            with mock.patch("os.listdir", return_value=[]):
                self._run_fingerprint_by_path()

        with pytest.raises(
            ClickException,
            match="Too many config files in the directory, only one has to present!",
        ):
            with mock.patch(
                "os.listdir",
                return_value=[
                    DEFAULT_CONNECTION_CONFIG_FILE,
                    DEFAULT_SKILL_CONFIG_FILE,
                ],
            ):
                self._run_fingerprint_by_path()


def _raise_exception(*args, **kwargs):
    raise Exception()


@mock.patch("aea.cli.fingerprint.open_file", mock.mock_open())
class FingerprintItemTestCase(TestCase):
    """Test case for fingerprint_item method."""

    @mock.patch("aea.cli.fingerprint.Path.exists", return_value=False)
    @mock.patch(
        "aea.cli.fingerprint.ConfigLoader.from_configuration_type",
        return_value=ConfigLoaderMock(),
    )
    def test_fingerprint_item_package_not_found(self, *mocks):
        """Test for fingerprint_item package not found result."""
        public_id = PublicIdMock()
        with self.assertRaises(ClickException) as cm:
            fingerprint_item(ContextMock(), "skill", public_id)
        self.assertIn("Package not found at path", cm.exception.message)

    @mock.patch(
        "aea.cli.fingerprint.ConfigLoader.from_configuration_type", _raise_exception
    )
    def test_fingerprint_item_exception(self, *mocks):
        """Test for fingerprint_item exception raised."""
        public_id = PublicIdMock()
        with self.assertRaises(ClickException):
            fingerprint_item(ContextMock(), "skill", public_id)


class TestFingerprintAgent(AEATestCaseEmpty):
    """Check fingerprint for agent."""

    def test_fingerprint(self):
        """Check fingerprint calculated and checked properly."""
        r = self.invoke("fingerprint")
        assert "calculated" in r.stdout

        click_context = MagicMock()
        click_context.obj = Context(self._get_cwd(), "", registry_path=None)
        click_context.obj.config["skip_consistency_check"] = True

        _check_aea_project([click_context], check_finger_prints=True)

        (Path(self._get_cwd()) / "some_file.txt").write_text("sdfds")
        with pytest.raises(
            ClickException, match=r"Fingerprints for package .* do not match"
        ):
            _check_aea_project([click_context], check_finger_prints=True)

        self.invoke("fingerprint")
        _check_aea_project([click_context], check_finger_prints=True)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2026 Valory AG
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests of the ethereum module."""

import logging
import time
from pathlib import Path
from typing import Dict, cast

import pytest
from aea_ledger_ethereum import EthereumApi, EthereumCrypto

from tests.conftest import ETHEREUM_PRIVATE_KEY_PATH, MAX_FLAKY_RERUNS, ROOT_DIR


@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
@pytest.mark.integration
@pytest.mark.ledger
def test_get_contract_instance(ethereum_testnet_config, ganache):
    """Test the get contract instance method."""
    ec = EthereumCrypto(private_key_path=ETHEREUM_PRIVATE_KEY_PATH)
    ethereum_api = EthereumApi(**ethereum_testnet_config)
    full_path = Path(ROOT_DIR, "tests", "data", "dummy_contract", "build", "some.json")
    contract_interface = ethereum_api.load_contract_interface(full_path)
    tx = ethereum_api.get_deploy_transaction(
        contract_interface,
        ec.address,
        value=0,
        max_priority_fee_per_gas=1000000000,
        max_fee_per_gas=1000000000,
    )
    tx = ethereum_api.get_deploy_transaction(
        contract_interface,
        ec.address,
        value=0,
        gas=1000000,
        max_priority_fee_per_gas=1000000000,
        max_fee_per_gas=1000000000,
    )
    tx_signed = ec.sign_transaction(tx)
    tx_digest = ethereum_api.send_signed_transaction(tx_signed)
    time.sleep(1)
    receipt = ethereum_api.get_transaction_receipt(tx_digest)
    erc1155_contract_address = cast(Dict, receipt)["contractAddress"]
    interface = {"abi": [], "bytecode": b""}
    instance = ethereum_api.get_contract_instance(
        contract_interface=interface,
        contract_address=erc1155_contract_address,
    )
    assert str(type(instance)) == "<class 'web3._utils.datatypes.Contract'>"
    instance = ethereum_api.get_contract_instance(
        contract_interface=interface,
    )
    assert (
        str(type(instance)) == "<class 'web3._utils.datatypes.PropertyCheckingFactory'>"
    )


@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
@pytest.mark.integration
@pytest.mark.ledger
def test_gas_station_strategy(ethereum_testnet_config, ganache):
    """Test the get contract instance method."""
    ec = EthereumCrypto(private_key_path=ETHEREUM_PRIVATE_KEY_PATH)

    ethereum_api = EthereumApi(**ethereum_testnet_config)
    full_path = Path(ROOT_DIR, "tests", "data", "dummy_contract", "build", "some.json")
    contract_interface = ethereum_api.load_contract_interface(full_path)
    tx = ethereum_api.get_deploy_transaction(
        contract_interface, ec.address, value=0, gas_price_strategy="gas_station"
    )
    assert all(
        [
            key in tx
            for key in ["gas", "chainId", "value", "nonce", "gasPrice", "data", "from"]
        ]
    )


@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
@pytest.mark.integration
@pytest.mark.ledger
def test_eip1559_strategy(ethereum_testnet_config, ganache):
    """Test the get contract instance method."""
    ec = EthereumCrypto(private_key_path=ETHEREUM_PRIVATE_KEY_PATH)

    ethereum_api = EthereumApi(**ethereum_testnet_config)
    full_path = Path(ROOT_DIR, "tests", "data", "dummy_contract", "build", "some.json")
    contract_interface = ethereum_api.load_contract_interface(full_path)
    tx = ethereum_api.get_deploy_transaction(
        contract_interface, ec.address, value=0, gas_price_strategy="eip1559"
    )
    logging.info(tx.keys())
    assert all(
        [
            key in tx
            for key in [
                "gas",
                "chainId",
                "value",
                "nonce",
                "maxFeePerGas",
                "maxPriorityFeePerGas",
                "data",
                "from",
            ]
        ]
    )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2026 Valory AG
#   Copyright 2018-2021 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests of the behaviour classes of the generic seller skill."""

# pylint: skip-file

import logging
from pathlib import Path
from typing import cast
from unittest.mock import patch

from aea.helpers.search.models import Description
from aea.test_tools.test_skill import BaseSkillTestCase

from packages.fetchai.protocols.oef_search.message import OefSearchMessage
from packages.fetchai.skills.generic_seller.behaviours import (
    GenericServiceRegistrationBehaviour,
    LEDGER_API_ADDRESS,
)
from packages.fetchai.skills.generic_seller.strategy import GenericStrategy
from packages.valory.protocols.ledger_api.message import LedgerApiMessage

PACKAGE_ROOT = Path(__file__).parent.parent


class TestSkillBehaviour(BaseSkillTestCase):
    """Test behaviours of generic seller."""

    path_to_skill = PACKAGE_ROOT
    is_agent_to_agent_messages = False

    @classmethod
    def setup_class(cls):
        """Setup the test class."""
        super().setup_class()
        cls.service_registration = cast(
            GenericServiceRegistrationBehaviour,
            cls._skill.skill_context.behaviours.service_registration,
        )
        cls.strategy = cast(GenericStrategy, cls._skill.skill_context.strategy)
        cls.logger = cls._skill.skill_context.logger

        cls.registration_message = OefSearchMessage(
            dialogue_reference=("", ""),
            performative=OefSearchMessage.Performative.REGISTER_SERVICE,
            service_description="some_service_description",
        )
        cls.registration_message.sender = str(cls._skill.skill_context.skill_id)
        cls.registration_message.to = cls._skill.skill_context.search_service_address

        cls.mocked_description = Description({"foo1": 1, "bar1": 2})

    def setup_method(self):
        """Setup"""
        super().teardown_method()
        self._init_service_registration = self.service_registration.__dict__.copy()
        self._strategy_kwargs = self.strategy.__dict__.copy()

    def teardown_method(self):
        """Teardown"""
        super().teardown_method()
        self.strategy.__dict__.update(self._strategy_kwargs)
        self.service_registration.__dict__.update(self._init_service_registration)

    def test_setup_is_ledger_tx(self):
        """Test the setup method of the service_registration behaviour where is_ledger_tx is True."""
        # setup
        self.strategy._is_ledger_tx = True
        mocked_description_1 = "some_description_1"

        # operation
        with patch.object(
            self.strategy,
            "get_location_description",
            return_value=mocked_description_1,
        ):
            with patch.object(
                self.service_registration.context.logger, "log"
            ) as mock_logger:
                self.service_registration.setup()

        # after
        self.assert_quantity_in_outbox(2)

        # message 1
        has_attributes, error_str = self.message_has_attributes(
            actual_message=self.get_message_from_outbox(),
            message_type=LedgerApiMessage,
            performative=LedgerApiMessage.Performative.GET_BALANCE,
            to=LEDGER_API_ADDRESS,
            sender=str(self.skill.skill_context.skill_id),
            ledger_id=self.strategy.ledger_id,
            address=self.skill.skill_context.agent_address,
        )
        assert has_attributes, error_str

        # _register_agent
        has_attributes, error_str = self.message_has_attributes(
            actual_message=self.get_message_from_outbox(),
            message_type=OefSearchMessage,
            performative=OefSearchMessage.Performative.REGISTER_SERVICE,
            to=self.skill.skill_context.search_service_address,
            sender=str(self.skill.skill_context.skill_id),
            service_description=mocked_description_1,
        )
        assert has_attributes, error_str

        mock_logger.assert_any_call(logging.INFO, "registering agent on SOEF.")

    def test_setup_not_is_ledger_tx(self):
        """Test the setup method of the service_registration behaviour: where is_ledger_tx is False."""
        # setup
        self.strategy._is_ledger_tx = False
        mocked_description_1 = "some_description_1"

        # operation
        with patch.object(
            self.strategy,
            "get_location_description",
            return_value=mocked_description_1,
        ):
            with patch.object(
                self.service_registration.context.logger, "log"
            ) as mock_logger:
                self.service_registration.setup()

        # after
        self.assert_quantity_in_outbox(1)

        # _register_agent
        has_attributes, error_str = self.message_has_attributes(
            actual_message=self.get_message_from_outbox(),
            message_type=OefSearchMessage,
            performative=OefSearchMessage.Performative.REGISTER_SERVICE,
            to=self.skill.skill_context.search_service_address,
            sender=str(self.skill.skill_context.skill_id),
            service_description=mocked_description_1,
        )
        assert has_attributes, error_str

        mock_logger.assert_any_call(logging.INFO, "registering agent on SOEF.")

    def test_act_i(self):
        """Test the act method of the service_registration behaviour where failed_registration_msg IS None."""
        # setup
        self.service_registration.failed_registration_msg = None

        # operation
        assert self.service_registration.act() is None

        # after
        self.assert_quantity_in_outbox(0)

    def test_act_ii(self):
        """Test the act method of the service_registration behaviour where failed_registration_msg is NOT None."""
        # setup
        self.service_registration.failed_registration_msg = self.registration_message

        # operation
        with patch.object(
            self.service_registration.context.logger, "log"
        ) as mock_logger:
            self.service_registration.act()

        # after
        self.assert_quantity_in_outbox(1)
        has_attributes, error_str = self.message_has_attributes(
            actual_message=self.get_message_from_outbox(),
            message_type=type(self.registration_message),
            performative=self.registration_message.performative,
            to=self.registration_message.to,
            sender=str(self.skill.skill_context.skill_id),
            service_description=self.registration_message.service_description,
        )
        assert has_attributes, error_str

        mock_logger.assert_any_call(
            logging.INFO,
            f"Retrying registration on SOEF. Retry {self.service_registration._nb_retries} out of {self.service_registration._max_soef_registration_retries}.",
        )
        assert self.service_registration.failed_registration_msg is None

    def test_act_iii(self):
        """Test the act method of the service_registration behaviour where failed_registration_msg is NOT None and max retries is reached."""
        # setup
        self.service_registration.failed_registration_msg = self.registration_message
        self.service_registration._max_soef_registration_retries = 2
        self.service_registration._nb_retries = 2

        # operation
        self.service_registration.act()

        # after
        self.assert_quantity_in_outbox(0)
        assert self.skill.skill_context.is_active is False

    def test_register_service(self):
        """Test the register_service method of the service_registration behaviour."""
        # operation
        with patch.object(
            self.strategy,
            "get_register_service_description",
            return_value=self.mocked_description,
        ):
            with patch.object(self.logger, "log") as mock_logger:
                self.service_registration.register_service()

        # after
        self.assert_quantity_in_outbox(1)

        message = self.get_message_from_outbox()
        has_attributes, error_str = self.message_has_attributes(
            actual_message=message,
            message_type=OefSearchMessage,
            performative=OefSearchMessage.Performative.REGISTER_SERVICE,
            to=self.skill.skill_context.search_service_address,
            sender=str(self.skill.skill_context.skill_id),
            service_description=self.mocked_description,
        )
        assert has_attributes, error_str
        mock_logger.assert_any_call(
            logging.INFO, "registering agent's service on the SOEF."
        )

    def test_register_genus(self):
        """Test the register_genus method of the service_registration behaviour."""
        # operation
        with patch.object(
            self.strategy,
            "get_register_personality_description",
            return_value=self.mocked_description,
        ):
            with patch.object(self.logger, "log") as mock_logger:
                self.service_registration.register_genus()

        # after
        self.assert_quantity_in_outbox(1)

        message = self.get_message_from_outbox()
        has_attributes, error_str = self.message_has_attributes(
            actual_message=message,
            message_type=OefSearchMessage,
            performative=OefSearchMessage.Performative.REGISTER_SERVICE,
            to=self.skill.skill_context.search_service_address,
            sender=str(self.skill.skill_context.skill_id),
            service_description=self.mocked_description,
        )
        assert has_attributes, error_str
        mock_logger.assert_any_call(
            logging.INFO, "registering agent's personality genus on the SOEF."
        )

    def test_register_classification(self):
        """Test the register_classification method of the service_registration behaviour."""
        # operation
        with patch.object(
            self.strategy,
            "get_register_classification_description",
            return_value=self.mocked_description,
        ):
            with patch.object(self.logger, "log") as mock_logger:
                self.service_registration.register_classification()

        # after
        self.assert_quantity_in_outbox(1)

        message = self.get_message_from_outbox()
        has_attributes, error_str = self.message_has_attributes(
            actual_message=message,
            message_type=OefSearchMessage,
            performative=OefSearchMessage.Performative.REGISTER_SERVICE,
            to=self.skill.skill_context.search_service_address,
            sender=str(self.skill.skill_context.skill_id),
            service_description=self.mocked_description,
        )
        assert has_attributes, error_str
        mock_logger.assert_any_call(
            logging.INFO, "registering agent's personality classification on the SOEF."
        )

    def test_teardown(self):
        """Test the teardown method of the service_registration behaviour."""
        # setup
        mocked_description_1 = "some_description_1"
        mocked_description_2 = "some_description_2"

        # operation
        with patch.object(
            self.strategy,
            "get_unregister_service_description",
            return_value=mocked_description_1,
        ):
            with patch.object(
                self.strategy,
                "get_location_description",
                return_value=mocked_description_2,
            ):
                with patch.object(
                    self.service_registration.context.logger, "log"
                ) as mock_logger:
                    self.service_registration.teardown()

        # after
        self.assert_quantity_in_outbox(2)

        # message 1
        has_attributes, error_str = self.message_has_attributes(
            actual_message=self.get_message_from_outbox(),
            message_type=OefSearchMessage,
            performative=OefSearchMessage.Performative.UNREGISTER_SERVICE,
            to=self.skill.skill_context.search_service_address,
            sender=str(self.skill.skill_context.skill_id),
            service_description=mocked_description_1,
        )
        assert has_attributes, error_str
        mock_logger.assert_any_call(logging.INFO, "unregistering service from SOEF.")

        # message 2
        has_attributes, error_str = self.message_has_attributes(
            actual_message=self.get_message_from_outbox(),
            message_type=OefSearchMessage,
            performative=OefSearchMessage.Performative.UNREGISTER_SERVICE,
            to=self.skill.skill_context.search_service_address,
            sender=str(self.skill.skill_context.skill_id),
            service_description=mocked_description_2,
        )
        assert has_attributes, error_str
        mock_logger.assert_any_call(logging.INFO, "unregistering agent from SOEF.")



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2026 Valory AG
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This test module contains integration tests for P2PLibp2p connection."""

# pylint: skip-file

import itertools
import logging
import os
import random
from dataclasses import dataclass
from pathlib import Path
from typing import List

import pytest
from hypothesis import given, settings
from hypothesis import strategies as st

from aea.mail.base import Envelope

from packages.fetchai.protocols.default.message import DefaultMessage
from packages.valory.connections.p2p_libp2p.tests.base import libp2p_log_on_failure_all
from packages.valory.connections.test_libp2p.tests.base import (
    BaseP2PLibp2pAEATestCaseMany,
    BaseP2PLibp2pTest,
    LOCALHOST,
    ports,
    wait_for_condition,
)
from packages.valory.connections.test_libp2p.tests.conftest import (
    ACNWithBootstrappedEntryNodes,
    NodeConfig,
    local_nodes,
    public_nodes,
)

AEA_DEFAULT_LAUNCH_TIMEOUT = 30
AEA_LIBP2P_LAUNCH_TIMEOUT = 30

skip_if_ci_marker = [
    pytest.Mark(
        name="skipif",
        args=(bool(os.environ.get("IS_CI_WORKFLOW")),),
        kwargs={"reason": "public ACN node tests flaky on CI"},
    )
]


@pytest.fixture
def maddrs(request):
    """Fixture for multi addresses."""
    return request.param


@pytest.fixture
def delegate_uris_public_keys(request):
    """Fixture for delegate uris and public keys."""
    return request.param


@pytest.mark.integration
@libp2p_log_on_failure_all
class Libp2pConnectionDHTRelay(BaseP2PLibp2pTest):
    """Test that public DHT's relay service is working properly"""

    def setup_method(self):
        """Setup test"""
        assert len(self.nodes) > 1, "Test requires at least 2 public DHT node"
        for node in self.nodes:
            for _ in range(2):  # make pairs
                self.make_connection(relay=False, entry_peers=[node.maddr])

    def teardown_method(self):
        """Teardown after test method"""
        self._disconnect()
        self.multiplexers.clear()
        self.log_files.clear()

    @property
    def pairs_with_same_entry_peers(self):
        """Multiplexer pairs with different entry peers"""
        return itertools.zip_longest(*[iter(self.multiplexers)] * 2)

    @property
    def pairs_with_different_entry_peers(self):
        """Multiplexer pairs with different entry peers"""
        return itertools.permutations(self.multiplexers[::2], 2)

    def test_connectivity(self):
        """Test connectivity."""
        assert self.all_connected

    def test_communication_direct(self):
        """Test direct communication through the same entry peer"""

        for mux_pair in self.pairs_with_same_entry_peers:
            sender, to = (c.address for m in mux_pair for c in m.connections)
            envelope = self.enveloped_default_message(to=to, sender=sender)
            mux_pair[0].put(envelope)
            delivered_envelope = mux_pair[1].get(block=True, timeout=30)
            assert self.sent_is_delivered_envelope(envelope, delivered_envelope)

    def test_communication_indirect(self):
        """Test indirect communication through another entry peer"""

        for mux_pair in self.pairs_with_different_entry_peers:
            sender, to = (c.address for m in mux_pair for c in m.connections)
            envelope = self.enveloped_default_message(to=to, sender=sender)
            mux_pair[0].put(envelope)
            delivered_envelope = mux_pair[1].get(block=True, timeout=30)
            assert self.sent_is_delivered_envelope(envelope, delivered_envelope)


class Libp2pConnectionDHTDelegate(Libp2pConnectionDHTRelay):
    """Test that public DHTs delegate service is working properly"""

    def setup_method(self):  # overwrite the setup, reuse the rest
        """Set up test"""

        assert len(self.nodes) > 1
        for node in self.nodes:
            for _ in range(2):
                self.make_client_connection(
                    uri=node.uri, peer_public_key=node.public_key
                )


@pytest.mark.integration
@libp2p_log_on_failure_all
class Libp2pConnectionDHTRelayAEACli(BaseP2PLibp2pAEATestCaseMany):
    """Test that public DHT's relay service is working properly, using aea cli"""

    def set_libp2p_connection_config(self):
        """Set libp2p connection config"""

        log_file = str(Path(f"libp2p_node_{self.agent_name}.log").absolute())
        self.log_files.append(log_file)
        self.nested_set_config(
            f"{self.p2p_libp2p_path}.config",
            {
                "local_uri": f"{LOCALHOST.netloc}:{next(ports)}",
                "entry_peers": [node.maddr for node in self.nodes],
                "log_file": log_file,
                "ledger_id": self.node_ledger_id,
            },
        )
        self.run_cli_command("issue-certificates", cwd=self._get_cwd())

    def test_connectivity(self):
        """Test connectivity."""

        self.add_libp2p_connection()
        self.set_libp2p_connection_config()
        process = self.run_agent()

        is_running = self.is_running(process, timeout=AEA_LIBP2P_LAUNCH_TIMEOUT)
        assert is_running, "AEA not running within timeout!"

        check_strings = "Peer running in "
        missing_strings = self.missing_from_output(process, check_strings)
        assert not missing_strings


@pytest.mark.integration
@libp2p_log_on_failure_all
class Libp2pConnectionDHTDelegateAEACli(BaseP2PLibp2pAEATestCaseMany):
    """Test that public DHT's delegate service is working properly, using aea cli"""

    def set_libp2p_client_connection_config(self):
        """Set libp2p client connection config"""

        path = self.p2p_libp2p_client_path
        nodes = [{"uri": n.uri, "public_key": n.public_key} for n in self.nodes]
        cert_requests = [self.make_node_cert_request(n.public_key) for n in self.nodes]
        self.nested_set_config(f"{path}.config", {"nodes": nodes})
        self.nested_set_config(f"{path}.cert_requests", cert_requests)
        self.run_cli_command("issue-certificates", cwd=self._get_cwd())

    def test_connectivity(self):
        """Test connectivity."""

        self.add_libp2p_client_connection()
        self.set_libp2p_client_connection_config()
        process = self.run_agent()
        is_running = self.is_running(process, timeout=AEA_DEFAULT_LAUNCH_TIMEOUT)
        assert is_running, "AEA not running within timeout!"


test_classes = [
    Libp2pConnectionDHTRelay,
    Libp2pConnectionDHTDelegate,
    Libp2pConnectionDHTRelayAEACli,
    Libp2pConnectionDHTDelegateAEACli,
]


@dataclass
class NodeConfigTestCase:
    """TestCase"""

    name: str
    nodes: List[NodeConfig]
    use_local_acn: bool = False


# dynamically create tests
for base_cls in test_classes:
    for test_case in (
        NodeConfigTestCase("Local", local_nodes, True),
        NodeConfigTestCase("Public", public_nodes),
    ):
        name = f"Test{test_case.name}{base_cls.__name__}"

        if test_case.use_local_acn:
            bases = base_cls, ACNWithBootstrappedEntryNodes
            test_cls = type(name, bases, {})
        else:
            test_cls = type(name, (base_cls,), {})
            test_cls.pytestmark = skip_if_ci_marker

        test_cls.__name__ = name
        test_cls.nodes = test_case.nodes
        globals()[test_cls.__name__] = test_cls


RANGE_32_BIT = -1 << 31, (1 << 31) - 1

base_message_strategy = dict(
    dialogue_reference=st.tuples(st.text(), st.text()),
    message_id=st.integers(*RANGE_32_BIT),
    target=st.integers(*RANGE_32_BIT),
)
default_message_strategy = dict(
    **base_message_strategy,
    content=st.binary(),
    performative=st.just(DefaultMessage.Performative.BYTES),
)


class TestDHTRobustness(BaseP2PLibp2pTest, ACNWithBootstrappedEntryNodes):
    """Test DHT Robustness"""

    nodes = local_nodes
    pytestmark = skip_if_ci_marker

    def setup_method(self):
        """Setup"""

        for node in self.nodes:
            self.make_connection(relay=False, entry_peers=[node.maddr])
            self.make_client_connection(uri=node.uri, peer_public_key=node.public_key)

    def teardown_method(self):
        """Teardown"""

        self._disconnect()
        self.multiplexers.clear()
        self.log_files.clear()

    @pytest.mark.parametrize("exponent", [2, 3, 4])
    def test_prolonged_message_exchange(self, exponent: int):
        """Test prolonged message exchange"""

        n_messages = 10**exponent
        for _ in range(n_messages):
            mux_pair = random.sample(self.multiplexers, 2)  # nosec - only for testing
            sender, to = (c.address for m in mux_pair for c in m.connections)
            envelope = self.enveloped_default_message(to=to, sender=sender)
            mux_pair[0].put(envelope)
            delivered_envelope = mux_pair[1].get(block=True, timeout=5)
            assert self.sent_is_delivered_envelope(envelope, delivered_envelope)

    @pytest.mark.parametrize("exponent", [2, 3, 4])
    def test_ship_first_check_later(self, exponent: int):
        """Ship first check later"""

        shipped, delivered = [], []
        n_messages = 10**exponent
        for _ in range(n_messages):
            mux_pair = random.sample(self.multiplexers, 2)  # nosec - only for testing
            sender, to = (c.address for m in mux_pair for c in m.connections)
            envelope = self.enveloped_default_message(to=to, sender=sender)
            mux_pair[0].put(envelope)
            shipped.append(envelope)

        def check():
            for mux in self.multiplexers:
                while not mux.in_queue.empty():
                    delivered.append(mux.get())
            logging.info(f"{len(delivered)} / {len(shipped)}")
            return len(delivered) >= n_messages

        timeout = max(2, n_messages // 10)
        try:
            wait_for_condition(lambda: check(), timeout, period=1.0)
        except TimeoutError:
            raise TimeoutError(f"Found only {len(delivered)} / {len(shipped)} messages")

        assert {e.encode() for e in shipped} == {e.encode() for e in delivered}

    def send_message_via_random_multiplexer_pair(self, message) -> bool:
        """Send message via random multiplexers"""

        mux_pair = random.sample(self.multiplexers, 2)  # nosec - only for testing
        sender, to = (c.address for m in mux_pair for c in m.connections)
        envelope = Envelope(
            to=to,
            sender=sender,
            protocol_specification_id=message.protocol_specification_id,
            message=message,
        )
        logging.debug(envelope)
        mux_pair[0].put(envelope)
        delivered_envelope = mux_pair[1].get(block=True, timeout=5)
        return self.sent_is_delivered_envelope(envelope, delivered_envelope)

    @settings(deadline=2000)
    @given(st.builds(DefaultMessage, **default_message_strategy))
    def test_randomized_default_message_exchange(self, message):
        """Test randomized default message strategy"""

        assert self.send_message_via_random_multiplexer_pair(message)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2026 Valory AG
#   Copyright 2018-2021 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Ledger TX generation and processing benchmark."""

import time
from contextlib import contextmanager
from typing import List

import docker
from aea_cli_benchmark.case_tx_generate.docker_image import (
    DockerImage,
    FetchLedgerDockerImage,
    GanacheDockerImage,
)
from cosmpy.clients.signing_cosmwasm_client import SigningCosmWasmClient
from cosmpy.common.rest_client import RestClient
from cosmpy.crypto.address import Address as CosmpyAddress
from cosmpy.crypto.keypairs import PrivateKey
from cosmpy.protos.cosmos.base.v1beta1.coin_pb2 import Coin

FETCHD_INITIAL_TX_SLEEP = 6
GAS_PRICE_API_KEY = ""
DEFAULT_FETCH_LEDGER_ADDR = "http://127.0.0.1"
DEFAULT_FETCH_LEDGER_REST_PORT = 1317
DEFAULT_GANACHE_ADDR = "http://127.0.0.1"
DEFAULT_GANACHE_PORT = 8545
DEFAULT_GANACHE_CHAIN_ID = 1337


FUNDED_ETH_PRIVATE_KEY_1 = (
    "0xa337a9149b4e1eafd6c21c421254cf7f98130233595db25f0f6f0a545fb08883"
)

DEFAULT_AMOUNT = 1000000000000000000000
DEFAULT_FETCH_MNEMONIC = "gap bomb bulk border original scare assault pelican resemble found laptop skin gesture height inflict clinic reject giggle hurdle bubble soldier hurt moon hint"
DEFAULT_MONIKER = "test-node"
DEFAULT_GENESIS_ACCOUNT = "validator"
DEFAULT_DENOMINATION = "atestfet"
DEFAULT_FETCH_CHAIN_ID = "stargateworld-3"

FETCHD_CONFIGURATION = dict(
    mnemonic=DEFAULT_FETCH_MNEMONIC,
    moniker=DEFAULT_MONIKER,
    chain_id=DEFAULT_FETCH_CHAIN_ID,
    genesis_account=DEFAULT_GENESIS_ACCOUNT,
    denom=DEFAULT_DENOMINATION,
)
FUNDED_FETCHAI_PRIVATE_KEY_1 = (
    "bbaef7511f275dc15f47436d14d6d3c92d4d01befea073d23d0c2750a46f6cb3"
)
GANACHE_CONFIGURATION = dict(
    accounts_balances=[(FUNDED_ETH_PRIVATE_KEY_1, DEFAULT_AMOUNT)]
)


DEFAULT_FETCH_LEDGER_RPC_PORT = 26657
DEFAULT_FETCH_DOCKER_IMAGE_TAG = "fetchai/fetchd:0.8.4"


@contextmanager
def _fetchd_context(fetchd_configuration, timeout: float = 2.0, max_attempts: int = 20):
    client = docker.from_env()
    image = FetchLedgerDockerImage(
        client,
        DEFAULT_FETCH_LEDGER_ADDR,
        DEFAULT_FETCH_LEDGER_RPC_PORT,
        DEFAULT_FETCH_DOCKER_IMAGE_TAG,
        config=fetchd_configuration,
    )
    yield from _launch_image(image, timeout=timeout, max_attempts=max_attempts)


def _launch_image(image: DockerImage, timeout: float = 2.0, max_attempts: int = 10):
    """
    Launch image.

    :param image: an instance of Docker image.
    :param timeout:  timeout to wait docker image launched.
    :param max_attempts: max attempts to check docker image is up.
    :yield: None
    """
    image.check_skip()
    image.stop_if_already_running()
    container = image.create()
    container.start()
    success = image.wait(max_attempts, timeout)
    if not success:
        container.stop()
        container.remove()
        raise Exception(f"{image.tag} doesn't work. Exiting...")
    else:
        try:
            time.sleep(timeout)
            yield
        finally:
            container.stop()
            container.remove()


@contextmanager
def _ganache_context(
    ganache_configuration,
    ganache_addr=DEFAULT_GANACHE_ADDR,
    ganache_port=DEFAULT_GANACHE_PORT,
    timeout: float = 2.0,
    max_attempts: int = 10,
):
    client = docker.from_env()
    image = GanacheDockerImage(
        client, ganache_addr, ganache_port, config=ganache_configuration
    )
    yield from _launch_image(image, timeout=timeout, max_attempts=max_attempts)


def fund_accounts_from_local_validator(
    addresses: List[str], amount: int, denom: str = DEFAULT_DENOMINATION
):
    """Send funds to local accounts from the local genesis validator."""
    rest_client = RestClient(
        f"{DEFAULT_FETCH_LEDGER_ADDR}:{DEFAULT_FETCH_LEDGER_REST_PORT}"
    )
    pk = PrivateKey(bytes.fromhex(FUNDED_FETCHAI_PRIVATE_KEY_1))

    time.sleep(FETCHD_INITIAL_TX_SLEEP)
    client = SigningCosmWasmClient(pk, rest_client, DEFAULT_FETCH_CHAIN_ID)
    coins = [Coin(amount=str(amount), denom=denom)]

    for address in addresses:
        client.send_tokens(CosmpyAddress(address), coins)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023-2026 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test `aea packages init` command."""

import click
import pytest

from aea.configurations.constants import PACKAGES
from aea.test_tools.test_cases import BaseAEATestCase


class TestInit(BaseAEATestCase):
    """Test init command."""

    use_packages_dir = False

    def test_init_repo(self) -> None:
        """Test init command."""

        result = self.run_cli_command(
            "--registry-path", str(self.t), "packages", "init"
        )

        assert result.exit_code == 0
        assert "Initialized packages repository" in result.stdout
        assert (self.t / PACKAGES).exists()


class TestInitFailure(BaseAEATestCase):
    """Test init command failure."""

    def test_init_repo(self) -> None:
        """Test init command."""

        with pytest.raises(
            click.ClickException, match="Packages repository already exists"
        ):
            self.run_cli_command("--registry-path", str(self.t), "packages", "init")



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2026 Valory AG
#   Copyright 2018-2021 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains tests for aea launcher."""

import shutil
import time
from multiprocessing import Event
from multiprocessing.managers import DictProxy  # type: ignore # noqa: F401
from pathlib import Path
from threading import Thread
from unittest.mock import patch

import pytest
import yaml

from aea.configurations.base import DEFAULT_AEA_CONFIG_FILE
from aea.launcher import AEADirMultiprocessTask, AEALauncher, _run_agent
from aea.test_tools.test_cases import AEATestCaseMany

from tests.common.utils import wait_for_condition
from tests.conftest import CUR_PATH, MAX_FLAKY_RERUNS

TIMEOUT = 30


@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
class TestThreadLauncherMode(AEATestCaseMany):
    """Test launcher in threaded mode."""

    RUNNER_MODE = "threaded"
    agent_name_1 = "myagent_1"
    agent_name_2 = "myagent_2"
    failing_agent = "failing_agent"

    @classmethod
    def setup_class(cls):
        """Set the test up."""
        super(AEATestCaseMany, cls).setup_class()
        cls.create_agents(cls.agent_name_1, cls.agent_name_2, cls.failing_agent)
        cls.set_agent_context(cls.failing_agent)
        shutil.copytree(
            Path(CUR_PATH, "data", "exception_skill"),
            Path(cls.t, cls.failing_agent, "skills", "exception"),
        )
        config_path = Path(cls.t, cls.failing_agent, DEFAULT_AEA_CONFIG_FILE)
        with open(config_path) as fp:
            config = yaml.safe_load(fp)
        config.setdefault("skills", []).append("fetchai/exception:0.1.0")
        yaml.safe_dump(config, open(config_path, "w"))
        cls.unset_agent_context()

        for agent_name in (cls.agent_name_1, cls.agent_name_2, cls.failing_agent):
            cls.set_agent_context(agent_name)
            cls.generate_private_key()
            cls.add_private_key()
            cls.set_runtime_mode_to_async(agent_name)
            cls.unset_agent_context()

    @classmethod
    def set_runtime_mode_to_async(cls, agent_name: str) -> None:
        """Set runtime mode of the agent to async."""
        config_path = Path(cls.t, agent_name, DEFAULT_AEA_CONFIG_FILE)
        with open(config_path) as fp:
            config = yaml.safe_load(fp)
        config.setdefault("runtime_mode", "async")
        with open(config_path, "w") as fp:
            yaml.safe_dump(config, fp)

    def test_start_stop(self, capfd, caplog) -> None:
        """Test agents started stopped."""
        try:
            runner = AEALauncher(
                [self.agent_name_1, self.agent_name_2],
                self.RUNNER_MODE,
                log_level="DEBUG",
            )
            runner.start(True)
            wait_for_condition(lambda: runner.is_running, timeout=10)

            capfd_out = ""

            def _check():
                nonlocal capfd_out  # to accumulate logs from capfd to chgck all the logs captured.
                capfd_out += capfd.readouterr().out
                log_text = capfd_out + "\n".join(caplog.messages)
                return (
                    f"[{self.agent_name_1}]: Runtime state changed to RuntimeStates.running."
                    in log_text
                    and f"[{self.agent_name_2}]: Runtime state changed to RuntimeStates.running."
                    in log_text
                )

            wait_for_condition(_check, timeout=TIMEOUT, period=0.1)
            assert runner.num_failed == 0
        finally:
            runner.stop()
            assert not runner.is_running
            assert runner.num_failed == 0

    def test_one_fails(self) -> None:
        """Test agents started, one agent failed, exception is raised."""
        try:
            runner = AEALauncher(
                [self.agent_name_1, self.agent_name_2, self.failing_agent],
                self.RUNNER_MODE,
            )

            with pytest.raises(Exception, match="Expected exception!"):
                runner.start()
        finally:
            runner.stop()

    def test_run_agent_in_thread(self):
        """Test agent started and stopped in thread."""
        stop_event = Event()
        t = Thread(target=_run_agent, args=(self.agent_name_1, stop_event))
        t.start()
        time.sleep(1)
        stop_event.set()
        t.join()


class TestAsyncLauncherMode(TestThreadLauncherMode):
    """Test launcher in async mode."""

    RUNNER_MODE = "async"


@pytest.mark.skip(
    reason="Unknown issue: https://github.com/valory-xyz/open-aea/actions/runs/7162860465/job/19500538744?pr=698"
)
class TestProcessLauncherMode(TestThreadLauncherMode):
    """Test launcher in process mode."""

    RUNNER_MODE = "multiprocess"


def test_task_stop():
    """Test AEADirMultiprocessTask.stop when not started."""
    task = AEADirMultiprocessTask("some")
    assert not task.failed
    with patch.object(task._stop_event, "set") as set_mock:
        task.stop()
        set_mock.assert_not_called()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2026 Valory AG
#   Copyright 2018-2021 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Multiagent manager utils."""

import datetime
import multiprocessing
import os
import sys
import time
import venv  # type: ignore
from traceback import format_exc
from typing import Any, Callable

from aea.crypto.plugin import load_all_plugins
from aea.manager.project import Project


def get_lib_path(env_dir: str) -> str:
    """Get librarty path for env dir."""
    if sys.platform == "win32":  # pragma: nocover
        libpath = os.path.join(env_dir, "Lib", "site-packages")
    else:  # pragma: nocover
        libpath = os.path.join(
            env_dir, "lib", "python%d.%d" % sys.version_info[:2], "site-packages"
        )
    return libpath


def make_venv(env_dir: str, set_env: bool = False) -> None:
    """
    Make venv and update variable to use it.

    :param env_dir: str, path for new env dir
    :param set_env: bool. use evn within this python process (update, sys.executable and sys.path)
    """
    builder = venv.EnvBuilder(True, clear=True)
    ctx = builder.ensure_directories(env_dir)
    builder.create(env_dir)
    if set_env:  # pragma: nocover
        sys.executable = ctx.env_exe
        sys.path.insert(0, get_lib_path(env_dir))


def project_install_and_build(project: Project) -> None:
    """Install project dependencies and build required components."""
    project.install_pypi_dependencies()
    load_all_plugins(is_raising_exception=False)
    project.build()


def get_venv_dir_for_project(project: Project) -> str:
    """Get virtual env directory for project specified."""
    return os.path.join(project.path, "venv")


def project_check(project: Project) -> None:
    """Perform project loads well."""
    project.check()


def run_in_venv(env_dir: str, fn: Callable, timeout: float, *args: Any) -> Any:
    """Run python function in a dedicated process with virtual env specified."""
    manager = multiprocessing.Manager()
    result_queue = manager.Queue()
    process = multiprocessing.Process(
        target=_run_in_venv_handler,
        args=tuple([env_dir, fn, result_queue] + list(args)),
    )
    process.start()
    start_time = time.time()
    while process.is_alive():
        time.sleep(0.005)
        if timeout != 0 and (time.time() - start_time > timeout):  # pragma: nocover
            process.terminate()
            process.join(5)
            raise TimeoutError()
    process.join(5)
    result = result_queue.get_nowait()
    if isinstance(result, BaseException):
        raise result
    return result


def _run_in_venv_handler(
    env_dir: str, fn: Callable, queue: multiprocessing.Queue, *args: Any
) -> None:
    """Do actual function run in a dedicated process within virtual env."""
    result = None
    try:
        make_venv(env_dir, set_env=True)
        result = fn(*args)
    except Exception as e:  # pylint: disable=broad-except
        print(
            f"Exception in venv runner at {datetime.datetime.now()} for {fn}:\n{format_exc()}"
        )
        result = e
    queue.put_nowait(result)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2026 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains tests for test case classes for AEA contract testing."""

from pathlib import Path
from typing import cast
from unittest import mock

import pytest
from aea_ledger_ethereum import EthereumApi

from aea.test_tools.test_contract import BaseContractTestCase
from aea.test_tools.utils import as_context, copy_class

from tests.data.dummy_contract.contract import DummyContract

DUMMY_TX = {"gasPrice": 0, "nonce": 0, "gas": 0}
TX_RECEIPT = {"raw_log": ""}
PATH_TO_DUMMY_CONTRACT = Path(*DummyContract.__module__.split(".")).parent.absolute()


# mocks
mock_get_deploy_transaction = mock.patch.object(
    EthereumApi, "get_deploy_transaction", return_value=DUMMY_TX
)
mock_send_signed_transaction = mock.patch.object(
    EthereumApi, "send_signed_transaction", return_value=""
)
mock_time_sleep = mock.patch("time.sleep", return_value=None)
mock_tx_receipt = mock.patch.object(
    EthereumApi, "get_transaction_receipt", return_value=TX_RECEIPT
)
mock_is_transaction_settled = mock.patch.object(
    EthereumApi, "is_transaction_settled", return_value=True
)
mock_get_balance_increment = mock.patch.object(
    EthereumApi, "get_balance", side_effect=range(4)
)


class TestCls(BaseContractTestCase):
    """Concrete `copy` of BaseContractTestCase"""

    path_to_contract = Path(".")
    ledger_identifier = ""

    @classmethod
    def finish_contract_deployment(cls) -> str:
        """Concrete finish_contract_deployment"""
        return ""


def test_base_contract_test_case_definition_without_attributes_raises_error() -> None:
    """Test that definition of concrete subclass of BaseContractTestCase without attributes raises error."""
    with pytest.raises(ValueError):

        class TestCls(BaseContractTestCase):
            pass

    with pytest.raises(ValueError):

        class TestClsB(BaseContractTestCase):
            path_to_contract = Path(".")


class TestBaseContractTestCaseSetup:
    """Test BaseContractTestCase setup."""

    def setup_method(self) -> None:
        """Setup test"""

        # must `copy` the class to avoid test interference
        self.test_cls = cast(TestCls, copy_class(TestCls))

    def setup_test_cls(self) -> TestCls:
        """Helper method to setup test to be tested"""

        test_instance = self.test_cls()  # type: ignore
        test_instance.setup_class()
        test_instance.setup_method()
        return test_instance

    def test_contract_setup_contract_configuration_not_found(self):
        """Test contract setup contract configuration not found"""

        self.test_cls.ledger_identifier = "ethereum"
        with pytest.raises(
            FileNotFoundError, match="Contract configuration not found: contract.yaml"
        ):
            self.setup_test_cls()

    def test_contract_setup_deploy_transaction_not_found(self):
        """Test contract setup deploy transaction not found"""

        self.test_cls.ledger_identifier = "ethereum"
        self.test_cls.path_to_contract = PATH_TO_DUMMY_CONTRACT
        with pytest.raises(ValueError, match="Deploy transaction not found!"):
            self.setup_test_cls()

    def test_contract_setup_transaction_digest_not_found(self):
        """Test contract setup transaction digest not found"""

        self.test_cls.ledger_identifier = "ethereum"
        self.test_cls.path_to_contract = PATH_TO_DUMMY_CONTRACT
        with as_context(
            mock_get_deploy_transaction,
            pytest.raises(ValueError, match="Transaction digest not found!"),
        ):
            self.setup_test_cls()

    def test_contract_setup_transaction_receipt_not_found(self):
        """Test contract setup transaction receipt not found"""

        self.test_cls.ledger_identifier = "ethereum"
        self.test_cls.path_to_contract = PATH_TO_DUMMY_CONTRACT
        with as_context(
            mock_get_deploy_transaction,
            mock_send_signed_transaction,
            mock_time_sleep,
            pytest.raises(ValueError, match="Transaction receipt not found!"),
        ):
            self.setup_test_cls()

    def test_contract_setup_transaction_receipt_not_valid(self):
        """Test contract setup transaction receipt not valid"""

        self.test_cls.ledger_identifier = "ethereum"
        self.test_cls.path_to_contract = PATH_TO_DUMMY_CONTRACT
        with as_context(
            mock_get_deploy_transaction,
            mock_send_signed_transaction,
            mock_time_sleep,
            mock_tx_receipt,
            pytest.raises(ValueError, match="Transaction receipt not valid!"),
        ):
            self.setup_test_cls()

    def test_contract_setup_successful(self):
        """Test contract setup successful"""

        self.test_cls.ledger_identifier = "ethereum"
        self.test_cls.path_to_contract = PATH_TO_DUMMY_CONTRACT
        with as_context(
            mock_get_deploy_transaction,
            mock_send_signed_transaction,
            mock_time_sleep,
            mock_tx_receipt,
            mock_is_transaction_settled,
        ):
            test = self.setup_test_cls()
            assert test.contract
            # exists for backward compatibility
            args = DUMMY_TX, test.ledger_api, test.deployer_crypto
            assert test.sign_send_confirm_receipt_transaction(*args) is TX_RECEIPT

    def test_contract_not_set(self):
        """Test contract not set"""

        test_instance = TestCls()
        with pytest.raises(
            ValueError, match="Ensure the contract is set during setup."
        ):
            assert test_instance.contract

    def test_fund_from_faucet_balance_not_increased(self):
        """Test fund fom faucet balance NOT increased"""

        self.test_cls.ledger_identifier = "ethereum"
        self.test_cls.path_to_contract = PATH_TO_DUMMY_CONTRACT
        self.test_cls.fund_from_faucet = True
        with as_context(
            mock_get_deploy_transaction,
            mock_send_signed_transaction,
            mock_time_sleep,
            mock_tx_receipt,
            mock_is_transaction_settled,
            pytest.raises(ValueError, match="Balance not increased!"),
        ):
            self.setup_test_cls()

    def test_fund_from_faucet_balance_increased(self):
        """Test fund fom faucet balance increased"""

        self.test_cls.ledger_identifier = "ethereum"
        self.test_cls.path_to_contract = PATH_TO_DUMMY_CONTRACT
        self.test_cls.fund_from_faucet = True
        with as_context(
            mock_get_deploy_transaction,
            mock_send_signed_transaction,
            mock_time_sleep,
            mock_tx_receipt,
            mock_is_transaction_settled,
            mock_get_balance_increment,
        ):
            self.setup_test_cls()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2026 Valory AG
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the tests of the ledger API connection module."""

# pylint: skip-file

import asyncio
import logging
import platform
import time
from pathlib import Path
from typing import Any, Dict, List, Optional, cast
from unittest.mock import Mock, patch

import pytest
from aea_ledger_ethereum import EthereumCrypto
from aea_ledger_ethereum.test_tools.constants import ETHEREUM_PRIVATE_KEY_PATH
from aea_ledger_ethereum.test_tools.fixture_helpers import (  # noqa: F401 pylint: disable=unsed-import
    DEFAULT_GANACHE_CHAIN_ID,
    ganache,
)
from eth_typing import BlockNumber
from web3.eth import Eth
from web3.types import FeeHistory, Wei

from aea.common import Address
from aea.configurations.data_types import PublicId
from aea.connections.base import Connection, ConnectionStates
from aea.crypto.ledger_apis import LedgerApis
from aea.crypto.registries import make_crypto, make_ledger_api
from aea.helpers.async_utils import AsyncState
from aea.helpers.transaction.base import (
    RawTransaction,
    SignedTransaction,
    Terms,
    TransactionDigest,
    TransactionReceipt,
)
from aea.mail.base import Envelope, Message
from aea.protocols.dialogue.base import Dialogue as BaseDialogue

from packages.valory.connections.ledger.connection import LedgerConnection
from packages.valory.connections.ledger.ledger_dispatcher import (
    LedgerApiRequestDispatcher,
)
from packages.valory.protocols.ledger_api.custom_types import (
    Kwargs,
    SignedTransactions,
    TransactionDigests,
)
from packages.valory.protocols.ledger_api.dialogues import (
    LedgerApiDialogue,
)
from packages.valory.protocols.ledger_api.dialogues import (
    LedgerApiDialogues as BaseLedgerApiDialogues,
)
from packages.valory.protocols.ledger_api.message import LedgerApiMessage

SOME_SKILL_ID = "some/skill:0.1.0"
PACKAGE_DIR = Path(__file__).parent.parent

skip_docker_tests = pytest.mark.skipif(
    platform.system() != "Linux",
    reason="Docker daemon is not available in Windows and macOS CI containers.",
)
logger = logging.getLogger(__name__)
ledger_ids = pytest.mark.parametrize(
    "ledger_id,address",
    [
        (EthereumCrypto.identifier, EthereumCrypto(ETHEREUM_PRIVATE_KEY_PATH).address),
    ],
)
# TODO: uncomment gas station strategy config after the gasstation API start
gas_strategies_parametrization = pytest.mark.parametrize(
    "gas_strategies",
    [
        {"gas_price_strategy": None},
        # {"gas_price_strategy": "gas_station"}, # noqa:E800
        {"gas_price_strategy": "eip1559"},
        {
            "max_fee_per_gas": 1_000_000_000,
            "max_priority_fee_per_gas": 1_000_000_000,
        },
    ],
)

fee_history_mock = patch("web3.eth.Eth.fee_history")

fee_history_parametrization = pytest.mark.parametrize(
    "fee_history",
    (
        FeeHistory(
            baseFeePerGas=[Wei(0)],
            gasUsedRatio=[0],
            oldestBlock=BlockNumber(0),
            reward=[[Wei(0)]],
        ),
        FeeHistory(
            baseFeePerGas=[
                Wei(int("0x180b5d502", 16)),
                Wei(int("0x18f9534de", 16)),
                Wei(int("0x18367e3a6", 16)),
                Wei(int("0x17e4519ed", 16)),
                Wei(int("0x179eb57e1", 16)),
                Wei(int("0x178ac8e88", 16)),
                Wei(int("0x177434776", 16)),
                Wei(int("0x17b4c8c2d", 16)),
                Wei(int("0x173695e84", 16)),
                Wei(int("0x1706c1f66", 16)),
                Wei(int("0x155600bc6", 16)),
            ],
            gasUsedRatio=[
                0.6546364333333333,
                0.37810266666666664,
                0.44697146666666665,
                0.4544759666666667,
                0.48681983333333334,
                0.4850136666666667,
                0.5430227,
                0.4168211333333333,
                0.46780666666666665,
                0.206346,
            ],
            oldestBlock=BlockNumber(int("0x13cd298", 16)),
            reward=[
                [Wei(int("0xf4240", 16))],
                [Wei(int("0xf4240", 16))],
                [Wei(int("0xf4240", 16))],
                [Wei(int("0x0", 16))],
                [Wei(int("0xf4240", 16))],
                [Wei(int("0xf4240", 16))],
                [Wei(int("0xf4240", 16))],
                [Wei(int("0x124f80", 16))],
                [Wei(int("0x919f8", 16))],
                [Wei(int("0xf4240", 16))],
            ],
        ),
        FeeHistory(
            baseFeePerGas=[],
            gasUsedRatio=[],
            oldestBlock=BlockNumber(0),
            reward=[],
        ),
    ),
)


class LedgerApiDialogues(BaseLedgerApiDialogues):
    """The dialogues class keeps track of all ledger_api dialogues."""

    def __init__(self, self_address: Address, **kwargs: Any) -> None:
        """Initialize dialogues."""

        def role_from_first_message(  # pylint: disable=unused-argument
            message: Message, receiver_address: Address
        ) -> BaseDialogue.Role:
            """Infer the role of the agent from an incoming/outgoing first message"""
            return LedgerApiDialogue.Role.AGENT

        BaseLedgerApiDialogues.__init__(
            self,
            self_address=self_address,
            role_from_first_message=role_from_first_message,
        )


@pytest.mark.usefixtures("ganache")
@skip_docker_tests
class TestLedgerDispatcher:
    """Tests for ledger dispatcher."""

    @pytest.mark.asyncio
    @ledger_ids
    async def test_get_balance(
        self,
        ledger_id: str,
        address: str,
        ledger_apis_connection: Connection,
        update_default_ethereum_ledger_api: None,
        ethereum_testnet_config: Dict,
    ) -> None:
        """Test get balance method."""

        config = ethereum_testnet_config
        ledger_api_dialogues = LedgerApiDialogues(SOME_SKILL_ID)
        request, ledger_api_dialogue = ledger_api_dialogues.create(
            counterparty=str(ledger_apis_connection.connection_id),
            performative=LedgerApiMessage.Performative.GET_BALANCE,  # type: ignore
            ledger_id=ledger_id,
            address=address,
        )
        envelope = Envelope(
            to=request.to,
            sender=request.sender,
            message=request,
        )

        await ledger_apis_connection.send(envelope)
        await asyncio.sleep(0.01)
        response = await ledger_apis_connection.receive()

        assert response is not None
        assert isinstance(response.message, LedgerApiMessage)
        response_msg = cast(LedgerApiMessage, response.message)
        response_dialogue = ledger_api_dialogues.update(response_msg)
        assert response_dialogue == ledger_api_dialogue
        assert response_msg.performative == LedgerApiMessage.Performative.BALANCE
        actual_balance_amount = response_msg.balance
        expected_balance_amount = make_ledger_api(ledger_id, **config).get_balance(
            address
        )
        assert actual_balance_amount == expected_balance_amount

    @pytest.mark.asyncio
    @ledger_ids
    async def test_get_state(
        self,
        ledger_id: str,
        address: str,
        ledger_apis_connection: Connection,
        update_default_ethereum_ledger_api: None,
        ethereum_testnet_config: Dict,
    ) -> None:
        """Test get state."""

        config = ethereum_testnet_config

        if "ethereum" in ledger_id:
            callable_name = "get_block"
        else:
            callable_name = "blocks"
        args = ("latest",)
        kwargs = Kwargs({})

        ledger_api_dialogues = LedgerApiDialogues(SOME_SKILL_ID)
        request, ledger_api_dialogue = ledger_api_dialogues.create(
            counterparty=str(ledger_apis_connection.connection_id),
            performative=LedgerApiMessage.Performative.GET_STATE,  # type: ignore
            ledger_id=ledger_id,
            callable=callable_name,
            args=args,
            kwargs=kwargs,
        )
        envelope = Envelope(
            to=request.to,
            sender=request.sender,
            message=request,
        )

        await ledger_apis_connection.send(envelope)
        await asyncio.sleep(0.01)
        response = await ledger_apis_connection.receive()

        assert response is not None
        assert isinstance(response.message, LedgerApiMessage)
        response_msg = cast(LedgerApiMessage, response.message)
        response_dialogue = ledger_api_dialogues.update(response_msg)
        assert response_dialogue == ledger_api_dialogue

        assert (
            response_msg.performative == LedgerApiMessage.Performative.STATE
        ), response_msg
        actual_block = response_msg.state.body
        expected_block = make_ledger_api(ledger_id, **config).get_state(
            callable_name, *args
        )
        assert actual_block == expected_block

    @fee_history_mock
    @pytest.mark.asyncio
    @gas_strategies_parametrization
    @fee_history_parametrization
    async def test_get_raw_transaction(
        self,
        fee_history_mock_: Mock,
        gas_strategies: Dict,
        fee_history: FeeHistory,
        ledger_apis_connection: Connection,
        update_default_ethereum_ledger_api: None,
    ) -> None:
        """Test get raw transaction with Ethereum APIs."""
        import aea  # noqa # to load registries

        fee_history_mock_.return_value = fee_history

        crypto1 = make_crypto(
            EthereumCrypto.identifier, private_key_path=ETHEREUM_PRIVATE_KEY_PATH
        )
        crypto2 = make_crypto(EthereumCrypto.identifier)
        ledger_api_dialogues = LedgerApiDialogues(SOME_SKILL_ID)

        amount = 40000
        fee = 10**7

        # Create ledger_api dialogue: get raw transaction
        request, ledger_api_dialogue = ledger_api_dialogues.create(
            counterparty=str(ledger_apis_connection.connection_id),
            performative=LedgerApiMessage.Performative.GET_RAW_TRANSACTION,  # type: ignore
            terms=Terms(
                ledger_id=EthereumCrypto.identifier,
                sender_address=crypto1.address,
                counterparty_address=crypto2.address,
                amount_by_currency_id={"ETH": -amount},
                quantities_by_good_id={"some_service_id": 1},
                is_sender_payable_tx_fee=True,
                nonce="",
                fee_by_currency_id={"ETH": fee},
                chain_id=DEFAULT_GANACHE_CHAIN_ID,
                **gas_strategies,
            ),
        )
        request = cast(LedgerApiMessage, request)
        envelope = Envelope(
            to=request.to,
            sender=request.sender,
            message=request,
        )
        await ledger_apis_connection.send(envelope)
        await asyncio.sleep(0.01)

        # Raw transaction
        response = await ledger_apis_connection.receive()

        assert response is not None
        assert isinstance(response.message, LedgerApiMessage)
        response_message = cast(LedgerApiMessage, response.message)
        assert (
            response_message.performative
            == LedgerApiMessage.Performative.RAW_TRANSACTION
        )
        response_dialogue = ledger_api_dialogues.update(response_message)
        assert response_dialogue == ledger_api_dialogue
        assert isinstance(response_message.raw_transaction, RawTransaction)
        assert response_message.raw_transaction.ledger_id == request.terms.ledger_id

    @fee_history_mock
    @pytest.mark.asyncio
    @gas_strategies_parametrization
    @fee_history_parametrization
    async def test_send_signed_transaction_ethereum(
        self,
        fee_history_mock_: Mock,
        gas_strategies: Dict,
        fee_history: FeeHistory,
        ledger_apis_connection: LedgerConnection,
    ) -> None:
        """Test send signed transaction with Ethereum APIs."""
        fee_history_mock_.return_value = fee_history
        ledger_api_dialogues = LedgerApiDialogues(SOME_SKILL_ID)

        crypto1 = make_crypto(
            EthereumCrypto.identifier, private_key_path=ETHEREUM_PRIVATE_KEY_PATH
        )
        crypto2 = make_crypto(EthereumCrypto.identifier)

        # First, send a transaction so we can get a digest at the end
        amount = 40000
        fee = 10**7

        request, ledger_api_dialogue = ledger_api_dialogues.create(
            counterparty=str(ledger_apis_connection.connection_id),
            performative=LedgerApiMessage.Performative.GET_RAW_TRANSACTION,  # type: ignore
            terms=Terms(
                ledger_id=EthereumCrypto.identifier,
                sender_address=crypto1.address,
                counterparty_address=crypto2.address,
                amount_by_currency_id={"ETH": -amount},
                quantities_by_good_id={"some_service_id": 1},
                is_sender_payable_tx_fee=True,
                nonce="",
                fee_by_currency_id={"ETH": fee},
                chain_id=DEFAULT_GANACHE_CHAIN_ID,
                **gas_strategies,
            ),
        )
        request = cast(LedgerApiMessage, request)
        envelope = Envelope(
            to=request.to,
            sender=request.sender,
            message=request,
        )

        # Check that we got the correct transaction response
        await ledger_apis_connection.send(envelope)
        await asyncio.sleep(0.01)
        response = await ledger_apis_connection.receive()

        assert response is not None
        assert isinstance(response.message, LedgerApiMessage)
        response_message = cast(LedgerApiMessage, response.message)
        assert (
            response_message.performative
            == LedgerApiMessage.Performative.RAW_TRANSACTION
        )
        response_dialogue = ledger_api_dialogues.update(response_message)
        assert response_dialogue == ledger_api_dialogue
        assert isinstance(response_message.raw_transaction, RawTransaction)
        assert response_message.raw_transaction.ledger_id == request.terms.ledger_id

        # Sign the transaction
        signed_transaction = crypto1.sign_transaction(
            response_message.raw_transaction.body
        )

        # Create new dialogue starting with signed transaction
        request, ledger_api_dialogue = ledger_api_dialogues.create(
            counterparty=str(ledger_apis_connection.connection_id),
            performative=LedgerApiMessage.Performative.SEND_SIGNED_TRANSACTION,  # type: ignore
            signed_transaction=SignedTransaction(
                EthereumCrypto.identifier, signed_transaction
            ),
        )
        request = cast(LedgerApiMessage, request)
        envelope = Envelope(
            to=request.to,
            sender=request.sender,
            message=request,
        )

        await ledger_apis_connection.send(envelope)
        await asyncio.sleep(0.01)

        # Transaction digest
        response = await ledger_apis_connection.receive()

        assert response is not None
        assert isinstance(response.message, LedgerApiMessage)
        response_message = cast(LedgerApiMessage, response.message)
        assert (
            response_message.performative != LedgerApiMessage.Performative.ERROR
        ), f"Received error: {response_message.message}"
        assert (
            response_message.performative
            == LedgerApiMessage.Performative.TRANSACTION_DIGEST
        )
        response_dialogue = ledger_api_dialogues.update(response_message)
        assert response_dialogue == ledger_api_dialogue
        assert isinstance(response_message.transaction_digest, TransactionDigest)
        assert isinstance(response_message.transaction_digest.body, str)
        assert (
            response_message.transaction_digest.ledger_id
            == request.signed_transaction.ledger_id
        )
        await asyncio.sleep(0.01)
        # Create new dialogue starting with GET_TRANSACTION_RECEIPT
        request, ledger_api_dialogue = ledger_api_dialogues.create(
            counterparty=str(ledger_apis_connection.connection_id),
            performative=LedgerApiMessage.Performative.GET_TRANSACTION_RECEIPT,  # type: ignore
            transaction_digest=response_message.transaction_digest,
        )
        envelope = Envelope(
            to=request.to,
            sender=request.sender,
            message=request,
        )
        await ledger_apis_connection.send(envelope)
        await asyncio.sleep(0.01)

        # Transaction receipt
        response = await ledger_apis_connection.receive()

        assert response is not None
        assert isinstance(response.message, LedgerApiMessage)
        response_message = cast(LedgerApiMessage, response.message)
        assert (
            response_message.performative
            == LedgerApiMessage.Performative.TRANSACTION_RECEIPT
        )
        response_dialogue = ledger_api_dialogues.update(response_message)
        assert response_dialogue == ledger_api_dialogue
        assert isinstance(response_message.transaction_receipt, TransactionReceipt)
        assert response_message.transaction_receipt.receipt is not None
        assert response_message.transaction_receipt.transaction is not None
        assert (
            response_message.transaction_receipt.ledger_id
            == request.transaction_digest.ledger_id  # type: ignore
        )
        assert LedgerApis.is_transaction_settled(
            response_message.transaction_receipt.ledger_id,
            response_message.transaction_receipt.receipt,
        ), "Transaction not settled."

    @pytest.mark.asyncio
    async def test_send_signed_transactions(
        self, ledger_apis_connection: LedgerConnection
    ) -> None:
        """Test the send_signed_transactions."""
        ledger_api_dialogues = LedgerApiDialogues(SOME_SKILL_ID)

        # Create new dialogue starting with signed transaction
        request, ledger_api_dialogue = ledger_api_dialogues.create(
            counterparty=str(ledger_apis_connection.connection_id),
            performative=LedgerApiMessage.Performative.SEND_SIGNED_TRANSACTIONS,  # type: ignore
            signed_transactions=SignedTransactions(
                EthereumCrypto.identifier,
                [{"raw_transaction": "test_tx"}],
            ),
            kwargs=LedgerApiMessage.Kwargs({}),
        )
        request = cast(LedgerApiMessage, request)
        envelope = Envelope(
            to=request.to,
            sender=request.sender,
            message=request,
        )

        expected_transaction_digests = ["transaction_digest_1", "transaction_digest_2"]
        # create a mock for api.send_signed_transactions
        with patch(
            "aea_ledger_ethereum.ethereum.EthereumApi.send_signed_transactions"
        ) as mock_send_signed_transactions:
            # configure the mock to return a specific value
            mock_send_signed_transactions.return_value = expected_transaction_digests

            # call the function under test
            await ledger_apis_connection.send(envelope)
            await asyncio.sleep(0.01)

            # Transaction digest
            response = await ledger_apis_connection.receive()

            assert response is not None
            assert isinstance(response.message, LedgerApiMessage)
            response_message = cast(LedgerApiMessage, response.message)
            assert (
                response_message.performative != LedgerApiMessage.Performative.ERROR
            ), f"Received error: {response_message.message}"
            assert (
                response_message.performative
                == LedgerApiMessage.Performative.TRANSACTION_DIGESTS
            )
            response_dialogue = ledger_api_dialogues.update(response_message)
            assert response_dialogue == ledger_api_dialogue
            assert isinstance(response_message.transaction_digests, TransactionDigests)
            assert isinstance(
                response_message.transaction_digests.transaction_digests, List
            )
            assert (
                response_message.transaction_digests.ledger_id
                == request.signed_transactions.ledger_id
            )
            assert (
                response_message.transaction_digests.transaction_digests
                == expected_transaction_digests
            )

    @pytest.mark.asyncio
    async def test_unsupported_protocol(
        self, ledger_apis_connection: LedgerConnection
    ) -> None:
        """Test fail on protocol not supported."""
        envelope = Envelope(
            to=str(ledger_apis_connection.connection_id),
            sender="test/skill:0.1.0",
            protocol_specification_id=PublicId.from_str("author/package_name:0.1.0"),
            message=b"message",
        )
        with pytest.raises(ValueError, match="Protocol not supported"):
            ledger_apis_connection._schedule_request(envelope)

    @pytest.mark.asyncio
    async def test_no_balance(
        self,
    ) -> None:
        """Test no balance."""
        dispatcher = LedgerApiRequestDispatcher(
            AsyncState(), connection_id=LedgerConnection.connection_id
        )
        mock_api = Mock()
        message = LedgerApiMessage(
            performative=LedgerApiMessage.Performative.GET_BALANCE,  # type: ignore
            dialogue_reference=dispatcher.dialogues.new_self_initiated_dialogue_reference(),
            ledger_id=EthereumCrypto.identifier,
            address="test",
        )
        message.to = dispatcher.dialogues.self_address
        message.sender = "test"
        dialogue = cast(
            Optional[LedgerApiDialogue], dispatcher.dialogues.update(message)
        )
        assert dialogue is not None

        mock_api.get_balance.return_value = None
        msg = dispatcher.get_balance(mock_api, message, dialogue)
        assert msg.performative == LedgerApiMessage.Performative.ERROR

    @pytest.mark.asyncio
    async def test_no_raw_tx(
        self,
    ) -> None:
        """Test no raw tx returned."""
        dispatcher = LedgerApiRequestDispatcher(
            AsyncState(), connection_id=LedgerConnection.connection_id
        )
        mock_api = Mock()
        message = LedgerApiMessage(
            performative=LedgerApiMessage.Performative.GET_RAW_TRANSACTION,  # type: ignore
            dialogue_reference=dispatcher.dialogues.new_self_initiated_dialogue_reference(),
            terms=Terms(
                ledger_id=EthereumCrypto.identifier,
                sender_address="1111",
                counterparty_address="22222",
                amount_by_currency_id={"ETH": -1},
                quantities_by_good_id={"some_service_id": 1},
                is_sender_payable_tx_fee=True,
                nonce="",
                fee_by_currency_id={"ETH": 10},
                chain_id=3,
            ),
        )
        message.to = dispatcher.dialogues.self_address
        message.sender = "test"
        dialogue = cast(
            Optional[LedgerApiDialogue], dispatcher.dialogues.update(message)
        )
        assert dialogue is not None

        mock_api.get_transfer_transaction.return_value = None
        msg = dispatcher.get_raw_transaction(mock_api, message, dialogue)
        assert msg.performative == LedgerApiMessage.Performative.ERROR

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "failing_ledger_method_name",
        ("get_transaction_receipt", "get_transaction"),
    )
    @pytest.mark.parametrize("retries", (0, 5, 20))
    @pytest.mark.parametrize("retry_timeout", (0.1,))
    @pytest.mark.parametrize("ledger_raise_error", (True, False))
    async def test_attempts_get_transaction_receipt(
        self,
        failing_ledger_method_name: str,
        retries: int,
        retry_timeout: float,
        ledger_raise_error: bool,
    ) -> None:
        """Test retry and sleep."""
        dispatcher = LedgerApiRequestDispatcher(
            AsyncState(ConnectionStates.connected),
            connection_id=LedgerConnection.connection_id,
        )
        mock_api = Mock()
        message = LedgerApiMessage(
            performative=LedgerApiMessage.Performative.GET_TRANSACTION_RECEIPT,  # type: ignore
            dialogue_reference=dispatcher.dialogues.new_self_initiated_dialogue_reference(),
            transaction_digest=TransactionDigest("asdad", "sdfdsf"),
        )
        message.to = dispatcher.dialogues.self_address
        message.sender = "test"
        dialogue = dispatcher.dialogues.update(message)
        assert dialogue is not None
        assert isinstance(dialogue, LedgerApiDialogue)

        failing_ledger_method = getattr(mock_api, failing_ledger_method_name)
        if ledger_raise_error:
            failing_ledger_method.side_effect = ValueError()
        else:
            failing_ledger_method.return_value = None

        with patch.object(dispatcher, "retry_attempts", retries):
            with patch.object(dispatcher, "retry_timeout", retry_timeout):
                msg = await dispatcher.get_transaction_receipt(
                    mock_api, message, dialogue
                )

        assert (
            msg.performative == LedgerApiMessage.Performative.ERROR
        ), "performative should be `ERROR`, please revisit the test's implementation."
        times_called = failing_ledger_method.call_count
        expected_times = retries
        assert times_called == expected_times, "Tried more times than expected!"

    @pytest.mark.asyncio
    @ledger_ids
    async def test_get_transaction_receipt_node_blocking(
        self,
        ledger_id: str,
        address: str,
        ledger_apis_connection: LedgerConnection,
        update_default_ethereum_ledger_api: None,
        ethereum_testnet_config: Dict,
    ) -> None:
        """Test retry strategy when the node is blocking."""
        retry_attempts = expected_times_called = 2
        retry_timeout = 0.001
        blocking_duration = 1

        # the retry strategy's total duration is an arithmetic progression
        expected_duration = sum(i * retry_timeout for i in range(retry_attempts))
        assert expected_duration < blocking_duration, (
            "The purpose of this test is to check whether the retry strategy works if a node is blocking."
            f"Therefore, the blocking time ({blocking_duration}) must be larger than the expected duration "
            f"({expected_duration}) of the retry strategy."
        )

        ledger_api_dialogues = LedgerApiDialogues(SOME_SKILL_ID)
        request, ledger_api_dialogue = ledger_api_dialogues.create(
            counterparty=str(ledger_apis_connection.connection_id),
            performative=LedgerApiMessage.Performative.GET_TRANSACTION_RECEIPT,  # type: ignore
            ledger_id=ledger_id,
            address=address,
            transaction_digest=TransactionDigest(ledger_id="ethereum", body="tx_hash"),
        )
        envelope = Envelope(
            to=request.to,
            sender=request.sender,
            message=request,
        )

        with patch.object(
            ledger_apis_connection._ledger_dispatcher, "retry_attempts", retry_attempts
        ), patch.object(
            ledger_apis_connection._ledger_dispatcher, "retry_timeout", retry_timeout
        ), patch.object(
            Eth,
            "get_transaction_receipt",
            side_effect=lambda *_: time.sleep(blocking_duration),
        ) as get_transaction_receipt_mock:
            await ledger_apis_connection.send(envelope)

            try:
                await asyncio.wait_for(
                    ledger_apis_connection.receive(), timeout=blocking_duration
                )
            except asyncio.exceptions.TimeoutError:
                raise AssertionError(
                    "The retry strategy did not finish before the given `blocking_duration`, "
                    "which suggests that the ledger api's call was also blocked, "
                    "and the dispatcher was waiting for its response."
                )

            actual_times_called = get_transaction_receipt_mock.call_count
            assert (
                actual_times_called == expected_times_called
            ), f"Tried {actual_times_called} times, {expected_times_called} were expected!"
